@@ -3,13 +3,13 @@
 
 mod error;
 pub mod load;
+pub mod tokens;
 
 use std::convert::TryInto;
 use std::path::Path;
 
-use reqwest::header;
+use reqwest::header::{self, HeaderMap};
 use reqwest::Client as HttpClient;
-use reqwest::ClientBuilder;
 use reqwest::{Body, RequestBuilder, StatusCode};
 use tokio_stream::{Stream, StreamExt};
 use tracing::{debug, info, instrument, trace};
@@ -27,80 +27,108 @@ pub type Result<T> = std::result::Result<T, ClientError>;
 pub const INVOICE_ENDPOINT: &str = "_i";
 pub const QUERY_ENDPOINT: &str = "_q";
 pub const RELATIONSHIP_ENDPOINT: &str = "_r";
+pub const LOGIN_ENDPOINT: &str = "login";
 const TOML_MIME_TYPE: &str = "application/toml";
 
 /// A client type for interacting with a Bindle server
 #[derive(Clone)]
-pub struct Client {
+pub struct Client<T> {
     client: HttpClient,
     base_url: Url,
+    token_manager: T,
 }
 
-/// Options for setting up a `Client`
-pub struct ClientOptions {
-    /// Controls whether the client assumes HTTP/2 or attempts to negotiate it.
-    /// The default is to assume HTTP/2, bypassing the negotiation step, because
-    /// the Bindle server always offers HTTP/2. However, it can be useful to negotiate
-    /// if the server may be behind a proxy or tunnelling service. Set
-    /// `http2_prior_knowledge` to `false` to negotiate.
-    pub http2_prior_knowledge: bool,
-    /// Controls whether the client accepts invalid certificates. The default is
-    /// to reject invalid certificates. It is sometimes necessary to set this
-    /// option in dev-test situations where you may be working with self-signed
-    /// certificates or the like.
-    pub danger_accept_invalid_certs: bool,
+/// The operation being performed against a Bindle server.
+enum Operation {
+    Create,
+    Yank,
+    Get,
+    Query,
+    Login,
 }
 
-impl Default for ClientOptions {
+/// A builder for for setting up a `Client`. Created using `Client::builder`
+pub struct ClientBuilder {
+    http2_prior_knowledge: bool,
+    danger_accept_invalid_certs: bool,
+}
+
+impl Default for ClientBuilder {
     fn default() -> Self {
         Self {
-            http2_prior_knowledge: true,
+            http2_prior_knowledge: false,
             danger_accept_invalid_certs: false,
         }
     }
 }
 
-impl Client {
-    /// Returns a new Client with the given URL, configured using the default options.
-    /// This URL should be the FQDN plus any namespacing
-    /// (like `v1`). So if you were running a bindle server mounted at the v1 endpoint, your URL
-    /// would look something like `http://my.bindle.com/v1/`. Will return an error if the URL is not
-    /// valid
-    pub fn new(base_url: &str) -> Result<Self> {
-        Self::new_with_options(base_url, ClientOptions::default())
+impl ClientBuilder {
+    /// Controls whether the client assumes HTTP/2 or attempts to negotiate it. Defaults to false.
+    pub fn http2_prior_knowledge(mut self, http2_prior_knowledge: bool) -> Self {
+        self.http2_prior_knowledge = http2_prior_knowledge;
+        self
     }
 
-    /// Returns a new Client with the given URL, configured using the given options.
-    /// This URL should be the FQDN plus any namespacing
-    /// (like `v1`). So if you were running a bindle server mounted at the v1 endpoint, your URL
-    /// would look something like `http://my.bindle.com/v1/`. Will return an error if the URL is not
-    /// valid
-    pub fn new_with_options(base_url: &str, options: ClientOptions) -> Result<Self> {
-        // Note that the trailing slash is important, otherwise the URL parser will treat is as a
-        // "file" component of the URL. So we need to check that it is added before parsing
-        let mut base = base_url.to_owned();
-        if !base.ends_with('/') {
-            info!("Provided base URL missing trailing slash, adding...");
-            base.push('/');
-        }
-        let base_parsed = Url::parse(&base)?;
-        let mut headers = header::HeaderMap::new();
-        headers.insert(header::ACCEPT, "application/toml".parse().unwrap());
-        // TODO: As this evolves, we might want to allow for setting time outs and accepting
-        // self-signed certs
+    /// Controls whether the client accepts invalid certificates. The default is to reject invalid
+    /// certificates. It is sometimes necessary to set this option in dev-test situations where you
+    /// may be working with self-signed certificates or the like. Defaults to false.
+    pub fn danger_accept_invalid_certs(mut self, danger_accept_invalid_certs: bool) -> Self {
+        self.danger_accept_invalid_certs = danger_accept_invalid_certs;
+        self
+    }
+
+    /// Returns a new Client with the given URL and token manager, configured using the set options.
+    ///
+    /// This URL should be the FQDN plus any namespacing (like `v1`). So if you were running a
+    /// bindle server mounted at the v1 endpoint, your URL would look something like
+    /// `http://my.bindle.com/v1/`. Will return an error if the URL is not valid
+    pub fn build<T>(self, base_url: &str, token_manager: T) -> Result<Client<T>> {
+        let (base_parsed, headers) = base_url_and_headers(base_url)?;
         let client = HttpClient::builder()
-            .and_if(options.http2_prior_knowledge, |b| b.http2_prior_knowledge())
-            .and_if(options.danger_accept_invalid_certs, |b| {
+            .and_if(self.http2_prior_knowledge, |b| b.http2_prior_knowledge())
+            .and_if(self.danger_accept_invalid_certs, |b| {
                 b.danger_accept_invalid_certs(true)
             })
             .default_headers(headers)
             .build()
             .map_err(|e| ClientError::Other(e.to_string()))?;
+
         Ok(Client {
             client,
             base_url: base_parsed,
+            token_manager,
         })
     }
+}
+
+pub(crate) fn base_url_and_headers(base_url: &str) -> Result<(Url, HeaderMap)> {
+    // Note that the trailing slash is important, otherwise the URL parser will treat is as a
+    // "file" component of the URL. So we need to check that it is added before parsing
+    let mut base = base_url.to_owned();
+    if !base.ends_with('/') {
+        info!("Provided base URL missing trailing slash, adding...");
+        base.push('/');
+    }
+    let base_parsed = Url::parse(&base)?;
+    let mut headers = header::HeaderMap::new();
+    headers.insert(header::ACCEPT, TOML_MIME_TYPE.parse().unwrap());
+    Ok((base_parsed, headers))
+}
+
+impl<T: tokens::TokenManager> Client<T> {
+    /// Returns a new Client with the given URL, configured using the default options.
+    ///
+    /// This URL should be the FQDN plus any namespacing (like `v1`). So if you were running a
+    /// bindle server mounted at the v1 endpoint, your URL would look something like
+    /// `http://my.bindle.com/v1/`. Will return an error if the URL is not valid
+    pub fn new(base_url: &str, token_manager: T) -> Result<Self> {
+        ClientBuilder::default().build(base_url, token_manager)
+    }
+
+    /// Returns a [`ClientBuilder`](ClientBuilder) configured with defaults
+    pub fn builder() -> ClientBuilder {
+        ClientBuilder::default()
+    }
 
     /// Performs a raw request using the underlying HTTP client and returns the raw response. The
     /// path is just the path part of your URL. It will be joined with the configured base URL for
@@ -116,6 +144,7 @@ impl Client {
         body: Option<impl Into<reqwest::Body>>,
     ) -> anyhow::Result<reqwest::Response> {
         let req = self.client.request(method, self.base_url.join(path)?);
+        let req = self.token_manager.apply_auth_header(req).await?;
         let req = match body {
             Some(b) => req.body(b),
             None => req,
@@ -132,7 +161,10 @@ impl Client {
         &self,
         inv: crate::Invoice,
     ) -> Result<crate::InvoiceCreateResponse> {
-        let req = self.create_invoice_builder().body(toml::to_vec(&inv)?);
+        let req = self
+            .create_invoice_builder()
+            .await?
+            .body(toml::to_vec(&inv)?);
         self.create_invoice_request(req).await
     }
 
@@ -150,15 +182,18 @@ impl Client {
         debug!("Successfully loaded invoice stream");
         let req = self
             .create_invoice_builder()
+            .await?
             .body(Body::wrap_stream(inv_stream));
         self.create_invoice_request(req).await
     }
 
-    fn create_invoice_builder(&self) -> RequestBuilder {
+    async fn create_invoice_builder(&self) -> Result<RequestBuilder> {
         // We can unwrap here because any URL error would be programmers fault
-        self.client
+        let req = self
+            .client
             .post(self.base_url.join(INVOICE_ENDPOINT).unwrap())
-            .header(header::CONTENT_TYPE, TOML_MIME_TYPE)
+            .header(header::CONTENT_TYPE, TOML_MIME_TYPE);
+        self.token_manager.apply_auth_header(req).await
     }
 
     async fn create_invoice_request(
@@ -167,7 +202,7 @@ impl Client {
     ) -> Result<crate::InvoiceCreateResponse> {
         trace!(?req);
         let resp = req.send().await?;
-        let resp = unwrap_status(resp, Endpoint::Invoice).await?;
+        let resp = unwrap_status(resp, Endpoint::Invoice, Operation::Create).await?;
         Ok(toml::from_slice(&resp.bytes().await?)?)
     }
 
@@ -210,9 +245,10 @@ impl Client {
 
     async fn get_invoice_request(&self, url: Url) -> Result<crate::Invoice> {
         let req = self.client.get(url);
+        let req = self.token_manager.apply_auth_header(req).await?;
         trace!(?req);
         let resp = req.send().await?;
-        let resp = unwrap_status(resp, Endpoint::Invoice).await?;
+        let resp = unwrap_status(resp, Endpoint::Invoice, Operation::Get).await?;
         Ok(toml::from_slice(&resp.bytes().await?)?)
     }
 
@@ -228,9 +264,10 @@ impl Client {
             .client
             .get(self.base_url.join(QUERY_ENDPOINT).unwrap())
             .query(&query_opts);
+        let req = self.token_manager.apply_auth_header(req).await?;
         trace!(?req);
         let resp = req.send().await?;
-        let resp = unwrap_status(resp, Endpoint::Query).await?;
+        let resp = unwrap_status(resp, Endpoint::Query, Operation::Query).await?;
         Ok(toml::from_slice(&resp.bytes().await?)?)
     }
 
@@ -252,9 +289,10 @@ impl Client {
             INVOICE_ENDPOINT,
             parsed_id.to_string()
         ))?);
+        let req = self.token_manager.apply_auth_header(req).await?;
         trace!(?req);
         let resp = req.send().await?;
-        unwrap_status(resp, Endpoint::Invoice).await?;
+        unwrap_status(resp, Endpoint::Invoice, Operation::Yank).await?;
         Ok(())
     }
 
@@ -278,6 +316,7 @@ impl Client {
         tracing::span::Span::current().record("invoice_id", &tracing::field::display(&parsed_id));
         self.create_parcel_request(
             self.create_parcel_builder(&parsed_id, parcel_sha)
+                .await?
                 .body(data),
         )
         .await
@@ -309,6 +348,7 @@ impl Client {
 
         self.create_parcel_request(
             self.create_parcel_builder(&parsed_id, parcel_sha)
+                .await?
                 .body(data_body),
         )
         .await
@@ -334,28 +374,34 @@ impl Client {
         let data_body = Body::wrap_stream(map);
         self.create_parcel_request(
             self.create_parcel_builder(&parsed_id, parcel_sha)
+                .await?
                 .body(data_body),
         )
         .await
     }
 
-    fn create_parcel_builder(&self, bindle_id: &Id, parcel_sha: &str) -> RequestBuilder {
+    async fn create_parcel_builder(
+        &self,
+        bindle_id: &Id,
+        parcel_sha: &str,
+    ) -> Result<RequestBuilder> {
         // We can unwrap here because any URL error would be programmers fault
-        self.client.post(
+        let req = self.client.post(
             self.base_url
                 .join(&format!(
                     "{}/{}@{}",
                     INVOICE_ENDPOINT, bindle_id, parcel_sha
                 ))
                 .unwrap(),
-        )
+        );
+        self.token_manager.apply_auth_header(req).await
     }
 
     async fn create_parcel_request(&self, req: RequestBuilder) -> Result<()> {
         // We can unwrap here because any URL error would be programmers fault
         trace!(?req);
         let resp = req.send().await?;
-        unwrap_status(resp, Endpoint::Parcel).await?;
+        unwrap_status(resp, Endpoint::Parcel, Operation::Create).await?;
         Ok(())
     }
 
@@ -403,9 +449,10 @@ impl Client {
                     .unwrap(),
             )
             .header(header::ACCEPT, "*/*");
+        let req = self.token_manager.apply_auth_header(req).await?;
         trace!(?req);
         let resp = req.send().await?;
-        unwrap_status(resp, Endpoint::Parcel).await
+        unwrap_status(resp, Endpoint::Parcel, Operation::Get).await
     }
 
     //////////////// Relationship Endpoints ////////////////
@@ -426,9 +473,10 @@ impl Client {
             "missing",
             parsed_id.to_string()
         ))?);
+        let req = self.token_manager.apply_auth_header(req).await?;
         trace!(?req);
         let resp = req.send().await?;
-        let resp = unwrap_status(resp, Endpoint::Invoice).await?;
+        let resp = unwrap_status(resp, Endpoint::Invoice, Operation::Get).await?;
         Ok(toml::from_slice::<crate::MissingParcelsResponse>(&resp.bytes().await?)?.missing)
     }
 }
@@ -436,7 +484,7 @@ impl Client {
 // We implement provider for client because often times (such as in the CLI) we are composing the
 // client in to a provider cache. This implementation does not verify or sign anything
 #[async_trait::async_trait]
-impl Provider for Client {
+impl<T: tokens::TokenManager + Send + Sync + 'static> Provider for Client<T> {
     async fn create_invoice<I>(
         &self,
         invoice: I,
@@ -542,17 +590,30 @@ enum Endpoint {
     Invoice,
     Parcel,
     Query,
+    // NOTE: This endpoint currently does nothing, but if we need more specific errors, we can use
+    // this down the line
+    Login,
 }
 
-async fn unwrap_status(resp: reqwest::Response, endpoint: Endpoint) -> Result<reqwest::Response> {
+async fn unwrap_status(
+    resp: reqwest::Response,
+    endpoint: Endpoint,
+    operation: Operation,
+) -> Result<reqwest::Response> {
     match (resp.status(), endpoint) {
         (StatusCode::OK, _) => Ok(resp),
         (StatusCode::ACCEPTED, Endpoint::Invoice) => Ok(resp),
         (StatusCode::CREATED, Endpoint::Invoice) => Ok(resp),
         (StatusCode::NOT_FOUND, Endpoint::Invoice) | (StatusCode::FORBIDDEN, Endpoint::Invoice) => {
-            Err(ClientError::InvoiceNotFound)
+            match operation {
+                Operation::Get => Err(ClientError::InvoiceNotFound),
+                _ => Err(ClientError::ResourceNotFound),
+            }
         }
-        (StatusCode::NOT_FOUND, Endpoint::Parcel) => Err(ClientError::ParcelNotFound),
+        (StatusCode::NOT_FOUND, Endpoint::Parcel) => match operation {
+            Operation::Get => Err(ClientError::ParcelNotFound),
+            _ => Err(ClientError::ResourceNotFound),
+        },
         (StatusCode::CONFLICT, Endpoint::Invoice) => Err(ClientError::InvoiceAlreadyExists),
         (StatusCode::CONFLICT, Endpoint::Parcel) => Err(ClientError::ParcelAlreadyExists),
         (StatusCode::UNAUTHORIZED, _) => Err(ClientError::Unauthorized),
@@ -596,4 +657,4 @@ trait ConditionalBuilder {
     }
 }
 
-impl ConditionalBuilder for ClientBuilder {}
+impl ConditionalBuilder for reqwest::ClientBuilder {}
